@@ -6,7 +6,8 @@ from collections import defaultdict
 from contextlib import contextmanager
 from functools import reduce
 from pathlib import Path
-from typing import Union, List, Tuple, Type, Dict, Set, Callable
+from textwrap import dedent
+from typing import Union, List, Tuple, Type, Dict, Set, Callable, Optional
 from uuid import uuid4
 
 import graphviz
@@ -21,7 +22,7 @@ from py2neo import ClientError, DatabaseError
 from tqdm import tqdm
 
 from .file import File, HDU
-from .graph import Graph
+from .graph import Graph, _convert_datatypes
 from .hierarchy import Multiple, Hierarchy, Graphable, OneOf
 from .path_finding import HierarchyGraph, get_all_class_bases
 from .readquery import Query
@@ -34,6 +35,10 @@ CONSTRAINT_FAILURE = re.compile(r"already exists with label `(?P<label>[^`]+)` a
                                 r"`(?P<idname>[^`]+)` = (?P<idvalue>[^`]+)$", flags=re.IGNORECASE)
 
 
+class ConnectionTimeOutError(ConnectionError):
+    pass
+
+
 def process_neo4j_error(data: 'Data', file: File, msg):
     matches = CONSTRAINT_FAILURE.findall(msg)
     if not len(matches):
@@ -243,7 +248,6 @@ class Data:
         self.password = password or os.getenv('WEAVEIO_PASSWORD', 'weavepassword')
         self.user = user or os.getenv('WEAVEIO_USER', 'weaveuser')
         self.rootdir = Path(rootdir or os.getenv('WEAVEIO_ROOTDIR', '/beegfs/car/weave/weaveio/'))
-
         self.write_allowed = False
         self.query = Query(self)
         self.rowparser = RowParser(self.rootdir)
@@ -408,30 +412,44 @@ class Data:
             d['user'] = self.user
         return Graph(host=self.host, port=self.port, name=self.dbname, write=self.write_allowed, **d)
 
-    def make_constraints_cypher(self):
+    def make_constraints_cypher(self) -> Dict[str, List[str]]:
         d = {hierarchy: hierarchy.make_schema() for hierarchy in self.hierarchies}
         d['temporary'] = ['CREATE CONSTRAINT ON (t:TemporaryMerge) ASSERT t.id IS UNIQUE']
+        d['dbcreated'] = ['CREATE INDEX n_dbcreated FOR (n:Hierarchy) ON n.`_dbcreated`']
+        d['dbupdated'] = ['CREATE INDEX n_dbupdated FOR (n:Hierarchy) ON n.`_dbupdated`']
         return d
 
     def apply_constraints(self):
         if not self.write_allowed:
             raise IOError(f"Writing is not allowed")
-        templates = []
-        equivalencies = []
+        templates = set()
+        equivalencies = set()
         for hier, qs in tqdm(self.make_constraints_cypher().items(), desc='applying constraints'):
             if not qs:
-                templates.append(hier)
+                templates.add(hier)
             for q in qs:
                 try:
                     self.graph.neograph.run(q)
                 except py2neo.ClientError as e:
                     if '[Schema.EquivalentSchemaRuleAlreadyExists]' in str(e):
-                       equivalencies.append(hier)
-                       templates.append(hier)
-        if len(templates):
-            logging.info(f'No index/constraint was made for {templates}')
+                        equivalencies.add(hier)
+                        templates.add(hier)
+        total = len(templates)
+        templates -= equivalencies
         if len(equivalencies):
-            logging.info(f'EquivalentSchemaRuleAlreadyExists for {equivalencies}')
+            n = f'({len(equivalencies)}/{total})'
+            equivalencies = ', '.join(sorted([getattr(t, '__name__', str(t)) for t in equivalencies]))
+            logging.info(f'EquivalentSchemaRuleAlreadyExists for {n}: {equivalencies}')
+        if len(templates):
+            n = f'({len(templates)}/{total})'
+            templates = ', '.join(sorted([getattr(t, '__name__', str(t)) for t in templates]))
+            logging.info(f'No index/constraint exists for {n}: {templates}')
+        self.await_indexes()
+
+    def await_indexes(self, timeout=300):
+        logging.info("Waiting for indexes to come online...")
+        self.graph.neograph.run(f'call db.awaitIndexes({timeout})')
+        logging.info("Indexes are online")
 
     def drop_all_constraints(self, indexes=True):
         if not self.write_allowed:
@@ -449,12 +467,13 @@ class Data:
         for index in tqdm(indexes, desc='dropping indexes'):
             self.graph.neograph.run(str(index)[1:-1])
 
-    def get_extant_files(self, *filetypes, return_only_fname=True):
+    def get_complete_files(self, *filetypes, return_only_fname=True):
         filetypes = [self.singular_hierarchies[self.singular_name(f)].__name__ for f in filetypes]
         if not filetypes:
             filetypes = ['File']
         r = 'fname' if return_only_fname else 'path'
-        result = self.graph.execute(f"MATCH (f:File) where any(l in labels(f) where l in $filetypes) RETURN DISTINCT f.{r}", filetypes=filetypes).to_series(dtype=str).values.tolist()
+        q = f"MATCH (f:File) where any(l in labels(f) where l in $filetypes) and f._dbcompleted_complete RETURN DISTINCT f.{r}"
+        result = self.graph.execute(q, filetypes=filetypes).to_series(0).values.tolist()
         if return_only_fname:
             return result
         return list(map(lambda x: self.rootdir / x, result))
@@ -481,10 +500,74 @@ class Data:
             raise IOError(f"Writing is not allowed")
         return self.graph.execute('match (n) where n.`_dbcreated` > $timestamp detach delete n', timestamp=state)
 
-    def write_files(self, *paths: Union[Path, str], raise_on_duplicate_file=False,
+    def mark_batch_complete_query(self, fname, slice, part, total_length, all_parts):
+        """
+        Mark fname[slice][part] as complete in the db and if all parts are then complete, mark that too
+        file._dbcompleted_{part} is a list of rows that have been marked as complete for that part
+        file._dbcompleted_{part}_complete = True if that part is completely written
+        file._dbcompleted_complete = True all parts are completely written
+        part is "None" for files which have no parts
+        """
+        def get_intervals(v):
+            return f"[c in apoc.coll.pairs([0] + [x in range(0, size({v})-2) where {v}[x] <> {v}[x+1] -1 | x+1]) | CASE WHEN c[1] is null THEN [{v}[c[0]],{v}[-1]] ELSE [{v}[c[0]], {v}[c[1]-1]] END]"
+        def get_list(v):
+            return f"apoc.coll.flatten([i in {v} | range(i[0],i[1])])"
+        q = dedent(rf"""
+        MATCH (f: File {{fname: $_check_fname}})
+        WITH *, apoc.coll.zip(f.`_dbcompleted_{part}_start`, f.`_dbcompleted_{part}_end`) as already_intervals
+        WITH *, {get_list('already_intervals')} as already_done
+        WITH *, apoc.coll.sort(apoc.coll.toSet(coalesce(already_done, []) + $_check_completed))+[-1] as done
+        WITH *, {get_intervals('done')} as done_intervals
+        WITH *, CASE WHEN size(done_intervals) = 0 THEN [[0, $_check_total_length-1]] ELSE done_intervals[..-1] END as done_intervals
+        SET f.`_dbcompleted_{part}_start` = [i in done_intervals | i[0]]
+        SET f.`_dbcompleted_{part}_end` = [i in done_intervals | i[1]]
+        SET f.`_dbcompleted_{part}_complete` = done_intervals = [[0, $_check_total_length-1]]
+        SET f.`_dbcompleted_complete` = all(x in $_check_parts where f['_dbcompleted_'+x+'_complete'])
+        """)
+        completed = list(range(0, total_length)[slice])
+        return q, {'_check_fname': fname, '_check_completed': completed,
+                   '_check_parts': list(map(str, all_parts)), '_check_total_length': total_length}
+
+    def file_batch_is_complete(self, filetype: File, path: Path, slc: slice, part: str) -> bool:
+        """
+        Returns True if the file[rows-slice][part] is complete otherwise False
+        """
+        total_length = filetype.length(self.rootdir / path, part)
+        q = dedent(f"""
+        OPTIONAL MATCH (f:File {{fname: $fname}})
+        with f, apoc.coll.zip(f._dbcompleted_{part}_start, f._dbcompleted_{part}_end) as intervals
+        return all(x in $rows where any(i in intervals where i[0] <= x and x <= i[1]))
+        """)
+        return self.graph.execute(q, fname=path.name, rows=list(range(total_length)[slc])).evaluate() or False
+
+    def files_to_batches(self, *paths: Union[Path, str], batch_size=None, batches_slc=None, parts=None,
+                         skip_complete=False):
+        batches = []
+        if not paths:
+            return []
+        if len(paths) == 1 and isinstance(paths[0], (tuple, list)):
+            paths = paths[0]
+        bar = tqdm(paths, desc='Building todo list')
+        for path in bar:
+            path = Path(path)
+            matches = [f for f in self.filetypes if f.match_file(self.rootdir, path.relative_to(self.rootdir), self.graph)]
+            if len(matches) > 1:
+                raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
+            filetype = matches[0]
+            filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
+            slices = filetype.get_batches(path, filetype_batch_size, parts, batches_slc)
+            bs = [(filetype, path.relative_to(self.rootdir), slc, part) for slc, part in slices]
+            if skip_complete:
+                bs = [b for b in bs if not self.file_batch_is_complete(*b)]
+            batches += bs
+        return batches
+
+    def write_files(self, *paths: Union[Path, str], raise_on_duplicate_file=False, skip_complete=True,
                     collision_manager='ignore', batch_size=None, parts=None, halt_on_error=True,
-                    dryrun=False, do_not_apply_constraints=False, test_one=False, infile_slc: slice = None,
-                    debug=False, debug_time=False) -> pd.DataFrame:
+                    dryrun=False, do_not_apply_constraints=False, test_one=False, batches_slc: slice = None,
+                    debug=False, debug_time=False, debug_params=False, debug_plan=False, timeout=None,
+                    replan='skip_except_first', runtime='pipelined',
+                    ) -> Optional[pd.DataFrame]:
         """
         Read in the files given in `paths` to the database.
         `collision_manager` is the method with which the database deals with overwriting data.
@@ -493,88 +576,118 @@ class Data:
         :return
             statistics dataframe
         """
+        if not halt_on_error:
+            logging.warning(f"`halt_on_error` is set to False. This would allow files to read without first reading their dependent files. "
+                            f"This is not recommended.")
         if not do_not_apply_constraints:
             self.apply_constraints()
-        batches = []
-        if len(paths) == 1 and isinstance(paths[0], (tuple, list)):
-            paths = paths[0]
-        for path in paths:
-            path = Path(path)
-            matches = [f for f in self.filetypes if f.match_file(self.rootdir, path.relative_to(self.rootdir), self.graph)]
-            if len(matches) > 1:
-                raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
-            filetype = matches[0]
-            filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
-            slices = filetype.get_batches(path, filetype_batch_size, parts, infile_slc)
-            batches += [(filetype, path.relative_to(self.rootdir), slc, part) for slc, part in slices]
+        batches = self.files_to_batches(*paths, batch_size=batch_size, batches_slc=batches_slc, parts=parts)
+        if not batches:
+            return
         elapsed_times = []
         stats = []
         timestamps = []
         if dryrun:
             logging.info(f"Dryrun: will not write to database. However, reading is permitted")
-        if test_one:
-            batches = batches[:1]
         _, path, slc, part = batches[0]
-        bar = tqdm(batches, desc=f'{path}[{slc.start}:{slc.stop}:{part}]')
+        desc_size = max(len(f'{path}[{slc.start}:{slc.stop}:{part}]') for filetype, path, slc, part in batches)
+        first = f"{path}[{slc.start}:{slc.stop}:{part}]"
+        bar = tqdm(batches, desc=first.ljust(desc_size))
         if debug or debug_time:
             with open('debug-timestamp.log', 'w') as f:
                 pass
-        for filetype, path, slc, part in bar:
+        out_batches = []
+        for i, (filetype, path, slc, part) in enumerate(bar):
             bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}]')
             try:
                 if raise_on_duplicate_file:
-                    if len(self.graph.execute('MATCH (f:File {fname: $fname})', fname=path.name)) != 0:
+                    if len(self.graph.execute('MATCH (f:File {fname: $fname}) return count(f)', fname=path.name)) != 0:
                         raise FileExistsError(f"{path.name} exists in the DB and raise_on_duplicate_file=True")
+                if skip_complete:
+                    if self.file_batch_is_complete(filetype, path, slc, part):
+                        continue
                 with self.write_cypher(collision_manager) as query:
                     filetype.read(self.rootdir, path, slc, part)
-                cypher, params = query.render_query()
-                uuid = f"//{uuid4()}"
-                cypher = '\n'.join([uuid, 'CYPHER runtime=interpreted', cypher])  # runtime is to avoid neo4j bug with pipelines: https://github.com/neo4j/neo4j/issues/12441
+                    cypher, params = query.render_query(as_lines=True)
+                    uuid = f"//{uuid4()}"
+                    guarantee, added_data = self.mark_batch_complete_query(path.name, slc, part, filetype.length(self.rootdir / path, part), filetype.parts)
+                    if replan == 'skip_except_first':
+                        replan_cypher = f'CYPHER replan=skip'
+                        replan = 'skip'
+                    else:
+                        replan_cypher = f'CYPHER replan={replan}'
+                    lines = [uuid, f'CYPHER runtime={runtime}', replan_cypher]  # runtime is to avoid neo4j bug with pipelines: https://github.com/neo4j/neo4j/issues/12441
+                    if debug_plan:
+                        lines.append('PROFILE')
+                    lines += cypher[:-1] + ['WITH time0', guarantee, cypher[-1]]
+                    params.update(added_data)
+                    cypher = '\n'.join(lines)
                 if debug:
                     with open('debug-query.log', 'w') as f:
                         f.write(cypher)
-                    with open('debug-params.log', 'w') as f:
-                        f.write(self.graph.output_for_debug(**params, silent=True))
+                    if debug_params:
+                        with open('debug-params.log', 'w') as f:
+                            f.write(self.graph.output_for_debug(**params, arrow=False, cmdline=False, silent=True))
                 start = time.time()
+                timed_out = False
                 if not dryrun:
                     try:
+                        self.graph.execute('MATCH (n) remove n._query_hash')
+                        self.graph.execute('MATCH ()-[r]-() remove r._query_hash')
                         results = self.graph.execute(cypher, **params)
                         stats.append(results.stats())
                         timestamp = results.evaluate()
+                        successful = True
                     except (ConnectionError, RuntimeError, IndexError, ConnectionResetError) as e:
                         is_running = True
+                        timed_out = False
                         while is_running:
                             is_running = self.graph.execute("CALL dbms.listQueries() YIELD query WHERE query STARTS WITH $uuid return count(*)", uuid=uuid).evaluate()
                             time.sleep(1)
-                            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}] (waiting for query to finish)')
+                            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}]')
+                            if timeout is not None:
+                                if time.time() - start > timeout:
+                                    timed_out = True
+                                    break
                         r = self.graph.execute('MATCH (f:File {fname: $fname}) return timestamp()', fname=path.name)
-                        successful = r.evaluate()
-                        timestamp = successful
-                        if not successful:
-                            raise ConnectionError(f"{path} could not be written to the database see neo4j logs for more details") from e
+                        timestamp = r.evaluate()
+                        successful = timestamp is not None
                         stats.append(r.stats())
+                        if not successful and halt_on_error:
+                            if timed_out:
+                                raise ConnectionTimeOutError(f"Connection timed out whilst writing {path}{slc.start}:{slc.stop}:{part} and it failed") from e
+                            else:
+                                raise ConnectionError(f"{path} could not be written to the database see neo4j logs for more details") from e
+                    finally:
+                        self.graph.execute('MATCH (n) remove n._query_hash')
+                        self.graph.execute('MATCH ()-[r]-() remove r._query_hash')
                     if timestamp is None:
                         logging.warning(f"This query terminated early due to either an empty input table/data or "
-                                        f"a match within the query returned no matches. "
+                                        f"a match within the query returned no matches or it timed-out. "
                                         f"Adjust your `.read` method and query to allow for empty tables/data")
                     timestamps.append(timestamp)
-                else:
-                    continue
-                    # self.graph.execute('EXPLAIN\n' + 'CYPHER runtime=interpreted\n' + cypher, **params)
-                elapsed_times.append(time.time() - start)
-                if debug or debug_time:
-                    with open('debug-timestamp.log', 'a') as f:
-                        f.write(str(elapsed_times[-1]) + '\n')
+                    if successful:
+                        elapsed_times.append(time.time() - start)
+                    else:
+                        elapsed_times.append(-1)
+                    out_batches.append((filetype, path, slc, part))
+                    if debug or debug_time:
+                        with open('debug-timestamp.log', 'a') as f:
+                            f.write(str(elapsed_times[-1]) + '\n')
             except (ClientError, DatabaseError, FileExistsError, ConnectionError) as e:
                 logging.exception('ClientError:', exc_info=True)
                 if halt_on_error:
                     raise e
                 print(e)
+            if test_one:
+                batches = batches[i:i+1]
+                logging.info(f"Writing terminated because `test_one=True`")
+                break
         if len(batches) and not dryrun:
             df = pd.DataFrame(stats)
             df['timestamp'] = timestamps
             df['elapsed_time'] = elapsed_times
-            _, df['fname'], slcs, parts = zip(*batches)
+            _, df['fname'], slcs, parts = zip(*out_batches)
             df['batch_start'], df['batch_end'] = zip(*[(i.start, i.stop) for i in slcs])
             df['part'] = parts
         elif dryrun:
@@ -587,102 +700,258 @@ class Data:
             return
         return df.set_index(['fname', 'batch_start', 'batch_end', 'part'])
 
-    def find_files(self, *filetype_names, skip_extant_files=True):
+    def find_files(self, *filetype_names, skip_complete_files=True):
+        filetype_names = [x+'_file' if not x.endswith('_file') else x for x in filetype_names]
+        if any(f for f in filetype_names if f not in [i.singular_name for i in self.filetypes]):
+            raise KeyError(f"Some or all of the filetype_names are not understood. "
+                           f"Allowed names are: {[i.singular_name.replace('_file', '') for i in self.filetypes]}")
         filelist = []
         if len(filetype_names) == 0:
             filetypes = self.filetypes
         else:
-            filetypes = [f for f in self.filetypes if f.singular_name in filetype_names or f.plural_name in filetype_names]
-        if len(filetypes) == 0:
-            raise KeyError(f"Some or all of the filetype_names are not understood. "
-                           f"Allowed names are: {[i.singular_name for i in self.filetypes]}")
+            filetypes = [f for f in self.filetypes if f.singular_name in filetype_names]
+
         for filetype in filetypes:
-            filelist += [i for i in filetype.match_files(self.rootdir, self.graph)]
-        if skip_extant_files:
-            extant_fnames = self.get_extant_files() if skip_extant_files else []
+            filelist += sorted([i for i in filetype.match_files(self.rootdir, self.graph)], key=lambda f: f.name)
+        if skip_complete_files:
+            extant_fnames = self.get_complete_files() if skip_complete_files else []
             filtered_filelist = [i for i in filelist if str(i.name) not in extant_fnames]
         else:
             filtered_filelist = filelist
         diff = len(filelist) - len(filtered_filelist)
         if diff:
-            print(f'Skipping {diff} extant files (use skip_extant_files=False to go over them again)')
+            print(f'Skipping {diff} complete files (use skip_complete_files=False to go over them again)')
         return [f for f in tqdm(filtered_filelist, desc='getting MOS files') if File.check_mos(f)]
 
     def write_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False,
-                        batch_size=None, parts=None, dryrun=False) -> pd.DataFrame:
+                        batch_size=None, parts=None, dryrun=False, **kwargs) -> pd.DataFrame:
         filtered_filelist = self.find_files(*filetype_names, skip_extant_files=skip_extant_files)
         return self.write_files(*filtered_filelist, collision_manager=collision_manager, halt_on_error=halt_on_error,
-                                dryrun=dryrun, batch_size=batch_size, parts=parts)
+                                dryrun=dryrun, batch_size=batch_size, parts=parts, **kwargs)
 
-    def _validate_one_required(self, hierarchy_name):
+    def _validate_one_required(self, hierarchy_name, timestamp=None):
         hierarchy = self.singular_hierarchies[hierarchy_name]
+        if hierarchy.is_template:
+            return
+        subclasses = get_all_subclasses(hierarchy)
         parents = [h for h in hierarchy.parents]
-        qs = []
-        for parent in parents:
-            if isinstance(parent, Multiple):
-                mn, mx = parent.minnumber, parent.maxnumber
-                b = parent.node.__name__
+        children = [h for h in hierarchy.children]
+        dfs = []
+        for relation in parents+children:
+            if isinstance(relation, Multiple):
+                mn, mx = relation.minnumber, relation.maxnumber
+                b = relation.node.__name__
+                reflected = relation.one2one
+                relation_idname = relation.relation_idname
             else:
                 mn, mx = 1, 1
-                b = parent.__name__
+                b = relation.__name__
+                reflected = False
+                relation_idname = None
             mn = 0 if mn is None else mn
             mx = 9999999 if mx is None else mx
             a = hierarchy.__name__
+            if relation in parents:
+                arrows = ['<-', '-']
+                reldef = [a, b]
+            else:
+                arrows = ['-', '->']
+                reldef = [b, a]
+            relid1 = f'WHERE r1.relation_id = "{relation_idname}"' if relation_idname else ''
+            relid2 = f'WHERE r2.relation_id = "{relation_idname}"' if relation_idname else ''
+            subclass_exclusions = ' WHERE ' + ' AND '.join([f'NOT n:{i.__name__}' for i in subclasses]) if len(subclasses) else ''
             q = f"""
-            MATCH (n:{a})
-            WITH n, SIZE([(n)<-[]-(m:{b}) | m ])  AS nodeCount
-            WHERE NOT (nodeCount >= {mn} AND nodeCount <= {mx})
-            RETURN "{a}", "{b}", {mn} as mn, {mx} as mx, n.id, nodeCount
+            MATCH (n:{a}){subclass_exclusions}
+            OPTIONAL MATCH (n){arrows[0]}[r1:is_required_by]{arrows[1]}(m:{b}) {relid1}
+            WITH n, count(r1) as forwardCount
+            OPTIONAL MATCH (m:{b}){arrows[0]}[r2:is_required_by]{arrows[1]}(n) {relid2}
+            WITH n, forwardCount, count(r2) as reflCount           
+            WITH *
+            WHERE forwardCount < {mn} OR forwardCount > {mx} OR (reflCount <> forwardCount AND {reflected})
+            RETURN "{reldef[0]}" as child, "{reldef[1]}" as parent, {mn} as mn, {mx} as mx, {reflected} as reflected, id(n), forwardCount, reflCount            
             """
-            qs.append(q)
-        if not len(parents):
-            qs = [f"""
-            MATCH (n:{hierarchy.__name__})
-            WITH n, SIZE([(n)<-[:IS_REQUIRED_BY]-(m) | m ])  AS nodeCount
-            WHERE nodeCount > 0
-            RETURN "{hierarchy.__name__}", "none", 0 as mn, 0 as mx, n.id, nodeCount
-            """]
-        dfs = []
-        for q in qs:
-            dfs.append(self.graph.neograph.run(q).to_data_frame())
-        df = pd.concat(dfs)
-        return df
+            result = self.graph.neograph.run(q).to_data_frame()
+            dfs.append(result)
+        try:
+            df = pd.concat(dfs)
+            return df, df.groupby(['child', 'parent']).apply(len).astype(int)
+        except KeyError:
+            return df, pd.Series()
+        except ValueError:
+            return pd.DataFrame(), pd.Series()
 
-    def _validate_no_duplicate_relation_ordering(self):
+    def _validate_no_duplicate_relation_ordering(self, time=None):
         q = """
         MATCH (a)-[r1]->(b)<-[r2]-(a)
         WHERE TYPE(r1) = TYPE(r2) AND r1.order <> r2.order
         WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
-        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
+        RETURN DISTINCT labels(a), id(a), labels(b), id(b), count(rs)+1
         """
         return self.graph.neograph.run(q).to_data_frame()
 
-    def _validate_no_duplicate_relationships(self):
+    def _validate_no_duplicate_relationships(self, timestamp=None):
         q = """
         MATCH (a)-[r1]->(b)<-[r2]-(a)
         WHERE TYPE(r1) = TYPE(r2) AND PROPERTIES(r1) = PROPERTIES(r2)
         WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
-        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
+        RETURN DISTINCT labels(a), id(a), labels(b), id(b), count(rs)+1
         """
         return self.graph.neograph.run(q).to_data_frame()
 
-    def validate(self):
-        duplicates = self._validate_no_duplicate_relationships()
-        print(f'There are {len(duplicates)} duplicate relations')
-        if len(duplicates):
-            print(duplicates)
-        duplicates = self._validate_no_duplicate_relation_ordering()
-        print(f'There are {len(duplicates)} relations with different orderings')
-        if len(duplicates):
-            print(duplicates)
+    def _validate_rel_unique_ids(self, *hiers):
+        """
+        For each hierarchy which defines an identifier builder based on parent hierarchies,
+        check that the parents+self is unique in the db since this is not guaranteed by neo4j indexes
+        """
+        instance_duplications = []
+        for h in tqdm(hiers or list(self.singular_hierarchies.keys()), desc='checking for schema uniqueness violations'):
+            hierarchy = self.singular_hierarchies[h]
+            if hierarchy.has_rel_identity() and not hierarchy.is_template:
+                id_factors = [f for f in hierarchy.factors if f in hierarchy.identifier_builder]
+                id_parents = [p if isinstance(p, Multiple) else OneOf(p) for p in hierarchy.parents]
+                id_children = [p if isinstance(p, Multiple) else OneOf(p) for p in hierarchy.children]
+                id_parents = {p.node.__name__: (p.minnumber, p.maxnumber) for p in id_parents if p.name in hierarchy.identifier_builder}
+                id_children = {p.node.__name__: (p.minnumber, p.maxnumber) for p in id_children if p.name in hierarchy.identifier_builder}
+                parent_nodes = [f"<-[rp{i}:is_required_by]-(p{i}:{p})" for i, p in enumerate(id_parents)]
+                parent_nodes2 = [f"<-[rp{i}2:is_required_by]-(p{i})" for i, p in enumerate(id_parents)]
+                child_nodes = [f"-[rc{i}:is_required_by]->(c{i}:{c})" for i, c in enumerate(id_children)]
+                child_nodes2 = [f"-[rc{i}2:is_required_by]->(c{i})" for i, c in enumerate(id_children)]
+                nodes = parent_nodes+child_nodes
+                nodes2 = parent_nodes2+child_nodes2
+                node = f"(n:{hierarchy.__name__})"
+                node2 = f"(m:{hierarchy.__name__})"
+                first = f"{node}{nodes[0]}"
+                first2 = f"{node2}{nodes2[0]}"
+                others = ','.join([f"(n){p}" for p in nodes[1:]])
+                others2 = ','.join([f"(m){p}" for p in nodes2[1:]])
+                parent_vars = ','.join([f"p{i}" for i in range(len(parent_nodes))]+[f"c{i}" for i in range(len(child_nodes))])
+                rel_vars = ','.join([f"rp{i}.order" for i in range(len(parent_nodes))]+[f"rc{i}.order" for i in range(len(child_nodes))])
+                rel_vars2 = ','.join([f"rp{i}2.order" for i in range(len(parent_nodes2))]+[f"rc{i}2.order" for i in range(len(child_nodes2))])
+                property_eq = ' and '.join([f"m.{f} = n.{f}" for f in id_factors])
+                if property_eq:
+                    property_eq = f' and {property_eq}'
+                if others:
+                    first = f"{first},{others}"
+                if others2:
+                    first2 = f"{first2},{others2}"
+                q = f"""
+                MATCH {first}
+                WITH n, {parent_vars}, [{rel_vars}] as coll1
+                OPTIONAL MATCH {first2}
+                where n <> m {property_eq} and [{rel_vars2}] = coll1
+                WITH n, count(m) as nduplicates, count(*) as cnt_rels
+                where nduplicates = cnt_rels  // if even 1 is different, then it's not a duplicate
+                return labels(n)[0] as type, id(n)
+                """
+                duplicates = self.graph.neograph.run(q).to_data_frame()
+                if len(duplicates):
+                    instance_duplications.append(duplicates)
+        try:
+            instance_duplications = pd.concat(instance_duplications)
+            class_duplications = instance_duplications.groupby('type').apply(len)
+            return instance_duplications, class_duplications
+        except ValueError:
+            return [], []
+
+
+    def _validate_expected_number_schema(self):
         schema_violations = []
-        for h in tqdm(list(self.singular_hierarchies.keys())):
-            schema_violations.append(self._validate_one_required(h))
+        label_instances = []
+        for h in tqdm(list(self.singular_hierarchies.keys()), desc='checking for schema parent/child # violations'):
+            df = self._validate_one_required(h)
+            if df is not None:
+                df, label_instance = df
+                schema_violations.append(df)
+                label_instances.append(label_instance)
         schema_violations = pd.concat(schema_violations)
-        print(f'There are {len(schema_violations)} violations of expected relationship number')
-        if len(schema_violations):
-            print(schema_violations)
-        return duplicates, schema_violations
+        label_instances = pd.concat(label_instances)
+        return schema_violations, label_instances
+
+    def _validate_relation_constraints(self):
+        qs = []
+        for h in tqdm(list(self.singular_hierarchies.keys()), desc='checking for rel constraint schema violations'):
+            hierarchy = self.singular_hierarchies[h]
+            if not hierarchy.is_template:
+                for rel in hierarchy.parents + hierarchy.children:
+                    if isinstance(rel, Multiple):
+                        for constraint in rel.constrain:
+                            q = f"""
+                            MATCH (h:{hierarchy.__name__})--(r:{rel.node.__name__})--(c:{constraint.__name__})
+                            WITH h, r, size(collect(c)) as size
+                            where size <> 1
+                            return h as node, r as dependent, "{constraint.__name__}" as constraint, size as ndifferent_constaints
+                            """
+                            qs.append(self.graph.neograph.run(q).to_data_frame())
+        return qs
+
+    def _validate_all_nodes_created_with_file(self):
+        q = dedent("""match (f:File) 
+        with collect(distinct f._dbcreated) as dbs
+        match (n) where not n.`_dbcreated` in dbs
+        with labels(n) as labels, count(n) as cnt, collect(id(n)) as ids
+        RETURN labels, cnt, ids""")
+        return self.graph.execute(q).to_data_frame()
+
+    def validate(self, timestamp: int = None, file_paths: List[Path]=None):
+        level = logging.getLogger().level
+        logging.getLogger().setLevel(logging.INFO)
+        missing = []
+        with pd.option_context('display.max_columns', None):
+            if file_paths is not None:
+                logging.info(f"Checking all requested files are present...")
+                extant = {path: fname for path, fname in self.graph.execute(f'MATCH (n:File) return n.path as path, n.fname as fname')}
+                missing = [str(path) for path in file_paths if path in extant]
+                if missing:
+                    logging.warning('fail')
+                    logging.warning(f"\tMissing {len(missing)} files from db")
+                    logging.warning(f"\tMissing the following files in the database:")
+                    logging.warning(f"{missing}")
+                else:
+                    logging.info('pass')
+            logging.info(f"Checking no duplicate relationships...")
+            duplicates = self._validate_no_duplicate_relationships(timestamp)
+            if len(duplicates):
+                logging.warning(f"fail")
+                logging.warning(f'\tThere are {len(duplicates)} duplicate relations')
+                logging.warning(f"\n{duplicates}")
+                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
+            else:
+                logging.info(f"pass")
+            logging.info(f"Checking no duplicate relationships with different ordering...")
+            duplicates = self._validate_no_duplicate_relation_ordering(timestamp)
+            if len(duplicates):
+                logging.warning(f"fail")
+                logging.warning(f'\tThere are {len(duplicates)} relations with different orderings')
+                logging.warning(f"\n{duplicates}")
+                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
+            else:
+                logging.info(f"pass")
+            logging.info(f"Checking expected number of parent/child relations from schema...")
+            schema_violations, label_instances = self._validate_expected_number_schema()
+            if len(schema_violations):
+                logging.warning(f"fail")
+                logging.warning(f'\tThere are {len(schema_violations)} violations of expected relationship number ({len(label_instances)} class relations)')
+                logging.warning(f"\n{schema_violations}")
+                logging.warning(f"\n{label_instances}")
+                logging.warning(f"\tThis indicates incorrect file read logic on user defined read functions. Check read function.")
+            else:
+                logging.info(f"pass")
+            logging.info(f"Checking unique relationships based on connected nodes are unique...")
+            unique_rel_id_duplicates, label_rel_id_duplicates = self._validate_rel_unique_ids()
+            if len(unique_rel_id_duplicates):
+                logging.warning(f"fail")
+                logging.warning(f"\tThere are {len(unique_rel_id_duplicates)} unique rel id violations ({len(label_rel_id_duplicates)} class definitions)")
+                logging.warning(f"\n{unique_rel_id_duplicates}")
+                logging.warning(f"\n{label_rel_id_duplicates}")
+                logging.warning(f"\tThis indicates a weaveio logic issue. Raise with maintainer.")
+            else:
+                logging.info(f"pass")
+            if len(missing) + len(duplicates) + len(schema_violations) + len(unique_rel_id_duplicates) == 0:
+                logging.info('Database schema is valid!')
+            else:
+                logging.warning(f"Database schema is invalid")
+            logging.getLogger().setLevel(level)
+            return duplicates, (schema_violations, label_instances), unique_rel_id_duplicates
 
     def is_product(self, factor_name, hierarchy_name):
         return self.singular_name(factor_name) in self.singular_hierarchies[self.singular_name(hierarchy_name)].products
@@ -773,6 +1042,7 @@ class Data:
     def is_valid_name(self, name):
         if not isinstance(name, str):
             return False
+        name = name.lower()
         if self.is_plural_name(name) or self.is_singular_name(name):
             return True
         if isinstance(name, str):
@@ -789,6 +1059,7 @@ class Data:
         """
         if not isinstance(name, str):
             return False
+        name = name.lower()
         if name in self.plural_hierarchies or name in self.plural_factors or\
                    name in self.plural_idnames or name in self.plural_relative_names:
             return True
@@ -801,6 +1072,7 @@ class Data:
     def is_singular_name(self, name):
         if not isinstance(name, str):
             return False
+        name = name.lower()
         if name in self.singular_hierarchies or name in self.singular_factors or \
            name in self.singular_idnames or name in self.relative_names:
             return True
