@@ -42,6 +42,8 @@ match_branch_node = hierarchy_query_decorator(writequery.match_branch_node)
 collect = hierarchy_query_decorator(writequery.collect)
 merge_relationship = hierarchy_query_decorator(writequery.merge_relationship)
 set_version = hierarchy_query_decorator(writequery.set_version)
+validate_number = hierarchy_query_decorator(writequery.validate_number)
+validate_type = hierarchy_query_decorator(writequery.validate_type)
 
 
 def chunker(lst, n):
@@ -66,27 +68,55 @@ def all_subclasses(cls):
 
 
 class Multiple:
-    def __init__(self, node, minnumber=1, maxnumber=None, constrain=None, idname=None, one2one=False, notreal=False):
+    def __init__(self, node: Union[str, 'GraphableMeta'], minnumber=1, maxnumber=None, constrain=None, idname=None, one2one=False, ordered=False, notreal=False):
+        """
+        Used to define relationships in the schema .e.g.:
+            class Car:
+                parents = [Multiple(Wheel, 4, 4, one2one=True), Multiple(Driver, constrain=(House,)]
+        This is read as:
+            A car has exactly 4 wheels and a wheel has only one car
+            A car has an unlimited number of drivers in the same house (but at least 1) and a driver has an unlimited number of cars (but at least 1)
+
+        :param node: A Hierarchy object to define a relationship to (e.g. creating a rel of Car-->Wheel
+        :param minnumber: The minimum number that can exists in the database, can be 0
+        :param maxnumber: The rough advised number of relationships that can exist
+        :param constrain: The Hierarchies which all node instances must have in their tree of dependencies
+        :param idname: The name of the relationship which can be used when reading
+        :param one2one: Is the relationship reciprocated? If yes, then a relationship of parent-[singular]->node  is also created
+                        e.g. After the Wheel-->Car rel is created, Car-[singular]->Wheel is also created.
+                            In this case a car has multiple wheels, but a wheel has only 1 car.
+        :param ordered: The order of the instances makes a different unique identification
+        :param notreal: This relationship is never instantiated. Use this to express logic about superclasses and subclasses
+        """
+        if not isinstance(node, (GraphableMeta, str)):
+            raise TypeError(f"{node} is of type {type(node)}. Must be a Hierarchy class or a name of a yet to be built Hierarchy class")
         self.node = node
         self.minnumber = int_or_none(minnumber) or 0
         if maxnumber is None:
             warn(f"maxnumber is None for {node}", RuntimeWarning)
         self.maxnumber = int_or_none(maxnumber)
-        self.constrain = [] if constrain is None else (constrain, ) if not isinstance(constrain, (list, tuple)) else tuple(constrain)
+        self.constrain = tuple() if constrain is None else (constrain, ) if not isinstance(constrain, (list, tuple)) else tuple(constrain)
         self.relation_idname = idname
         self.one2one = one2one
         self._isself = self.node == 'self'
         self.notreal = notreal
+        self.ordered = bool(ordered)
         if inspect.isclass(self.node):
             if issubclass(self.node, Hierarchy):
                 self.instantate_node()
 
+    def to_reldict(self):
+        return {'relation_idname': self.relation_idname, 'maxnumber': self.maxnumber, 'minnumber': self.minnumber,
+                'constrain': self.constrain, 'one2one': self.one2one, 'notreal': self.notreal, 'ordered': self.ordered}
+
     @property
     def is_optional(self):
         return self.minnumber == 0
 
     @property
     def name(self):
+        if self.relation_idname is not None:
+            return self.relation_idname
         if self.maxnumber == 1:
             return self.singular_name
         return self.plural_name
@@ -120,14 +150,18 @@ class Multiple:
             h.instantate_node(include_hierarchies)
 
     def __repr__(self):
-        return f"<Multiple({self.node} [{self.minnumber} - {self.maxnumber}] id={self.relation_idname})>"
+        if self.ordered:
+            o = '[Ordered]'
+        else:
+            o = '[Unordered]'
+        return f"<Multiple({self.node} [{self.minnumber} - {self.maxnumber}] id={self.relation_idname} {o})>"
 
     def __hash__(self):
         if isinstance(self.node, str):
             hsh = hash(self.node)
         else:
             hsh = hash(self.node.__name__)
-        return hash(self.__class__) ^ hash(self.minnumber) ^ hash(self.maxnumber) ^\
+        return hash(self.__class__) ^ hash(self.minnumber) ^ hash(self.maxnumber) ^ hash(self.ordered) ^ \
         reduce(lambda x, y: x ^ y, map(hash, self.constrain), 0) ^ hash(self.relation_idname) ^ hsh
 
     def __eq__(self, other):
@@ -139,6 +173,13 @@ class Multiple:
         multiple_list = [Multiple(hierarchy, i, i) if isinstance(i, int) else Multiple(cls, *i) for k, i in multiples.items()]
         return single_list + multiple_list
 
+    @classmethod
+    def from_any(cls, anything: Union[str, Type['Hierarchy'], 'Multiple']) -> 'Multiple':
+        if isinstance(anything, (str, GraphableMeta)):
+            return OneOf(anything)
+        elif isinstance(anything, Multiple):
+            return anything
+        raise TypeError(f"{anything} is not a subclass of Hierarchy or an instance of Multiple")
 
 class OneOf(Multiple):
     def __init__(self, node, constrain=None, idname=None, one2one=False):
@@ -147,10 +188,6 @@ class OneOf(Multiple):
     def __repr__(self):
         return f"<OneOf({self.node})>"
 
-    @property
-    def name(self):
-        return self.singular_name
-
 
 class Optional(Multiple):
     def __init__(self, node, constrain=None, idname=None, one2one=False):
@@ -159,10 +196,6 @@ class Optional(Multiple):
     def __repr__(self):
         return f"<Optional({self.node})>"
 
-    @property
-    def name(self):
-        return self.singular_name
-
 
 class GraphableMeta(type):
     def __new__(meta, name: str, bases, _dct):
@@ -206,10 +239,11 @@ class GraphableMeta(type):
         return r
 
     def __init__(cls, name, bases, dct):
+        invalid_bases = [b for b in bases if not b.is_template]
+        if any(invalid_bases):
+            raise RuleBreakingException(f"{name} cannot subclass {invalid_bases} since they are marked with `is_template=False`")
         if cls.idname is not None and cls.identifier_builder is not None:
             raise RuleBreakingException(f"You cannot define a separate idname and an identifier_builder at the same time for {name}")
-        if cls.indexes and (cls.idname is not None or cls.identifier_builder is not None):
-            raise RuleBreakingException(f"You cannot define an index and an id at the same time for {name}")
         parentnames = {}
         cls.children = deepcopy(cls.children)  # sever link so that changes here dont affect base classes
         cls.parents = deepcopy(cls.parents)
@@ -221,10 +255,7 @@ class GraphableMeta(type):
                 for n in c.constrain:
                     if n not in cls.children:
                         cls.children.append(n)
-                if c.maxnumber == 1:
-                    parentnames[c.singular_name] = (c.minnumber, c.maxnumber)
-                else:
-                    parentnames[c.plural_name] = (c.minnumber, c.maxnumber)
+                parentnames[c.name] = (c.minnumber, c.maxnumber)
             else:
                 parentnames[c.singular_name] = (1, 1)
         for i, p in enumerate(cls.parents):
@@ -235,10 +266,7 @@ class GraphableMeta(type):
                 for n in p.constrain:
                     if n not in cls.parents:
                         cls.parents.append(n)
-                if p.maxnumber == 1:
-                    parentnames[p.singular_name] = (p.minnumber, p.maxnumber)
-                else:
-                    parentnames[p.plural_name] = (p.minnumber, p.maxnumber)
+                parentnames[p.name] = (p.minnumber, p.maxnumber)
             else:
                 parentnames[p.singular_name] = (1, 1)
         if cls.identifier_builder is not None:
@@ -286,7 +314,7 @@ class GraphableMeta(type):
             if isinstance(relative, Multiple):
                 if relative.relation_idname is not None:
                     cls.relative_names[relative.relation_idname] = relative
-
+        cls.indexable = cls.idname is not None or cls.identifier_builder is not None
         super().__init__(name, bases, dct)
 
 
@@ -294,6 +322,7 @@ class Graphable(metaclass=GraphableMeta):
     idname = None
     identifier = None
     indexer = None
+    indexable = False
     type_graph_attrs = {}
     plural_name = None
     singular_name = None
@@ -387,8 +416,13 @@ class Graphable(metaclass=GraphableMeta):
 
 
     def __init__(self, predecessors, successors=None, do_not_create=False):
+        if self.is_template and not do_not_create:
+            raise TypeError(f"Template Hierarchies cannot be instantiated directly. "
+                            f"If you need to instantiate {self.__class__.__name__}, set {self.__class__.__name__}.is_template=False.")
         if successors is None:
             successors = {}
+        if predecessors is None:
+            predecessors = {}
         self.predecessors = predecessors
         self.successors = successors
         self.data = None
@@ -402,62 +436,135 @@ class Graphable(metaclass=GraphableMeta):
         merge_strategy = self.__class__.merge_strategy()
         version_parents = []
         if  merge_strategy == 'NODE FIRST':
-            self.node = child = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
+            self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
                                            collision_manager=collision_manager)
-            for k, parent_list in predecessors.items():
+            for k, (parent_list, relation) in predecessors.items():
                 type = 'is_required_by'
                 if isinstance(parent_list, Collection):
                     with unwind(parent_list, enumerated=True) as (parent, i):
-                        props = {'order': i, 'relation_id': k}
-                        merge_relationship(parent, child, type, {}, props, collision_manager=collision_manager)
+                        props = {'relation_id': k}
+                        if relation.ordered:
+                            props['order'] = i
+                        merge_relationship(parent, self.node, type, {}, props, collision_manager=collision_manager)
                     parent_list = collect(parent)
                     if k in self.version_on:
                         raise RuleBreakingException(f"Cannot version on a collection of nodes")
                 else:
-                    for parent in parent_list:
-                        props = {'order': 0, 'relation_id': k}
-                        merge_relationship(parent, child, type, {}, props, collision_manager=collision_manager)
+                    for i, parent in enumerate(parent_list):
+                        props = {'relation_id': k}
+                        if relation.ordered:
+                            props['order'] = i
+                        merge_relationship(parent, self.node, type, {}, props, collision_manager=collision_manager)
                         if k in self.version_on:
                             version_parents.append(parent)
-        elif merge_strategy == 'NODE+RELATIONSHIP':
-            parentnames = [p.name for p in self.parents]
-            parents = []
-            others = []
-            for k, parent_list in predecessors.items():
-                if isinstance(parent_list, Collection):
-                    raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
-                if k in parentnames and k in self.identifier_builder:
-                    parents += [p for p in parent_list]
+            # now the children
+            for k, (child_list, relation) in successors.items():
+                type = 'is_required_by'
+                if isinstance(child_list, Collection):
+                    with unwind(child_list, enumerated=True) as (child, i):
+                        props = {'relation_id': k}
+                        if relation.ordered:
+                            props['order'] = i
+                        merge_relationship(self.node, child, type,
+                                           {}, props,
+                                           collision_manager=collision_manager)
+                    collect(child)
                 else:
-                    others += [(i, k, p) for i, p in enumerate(parent_list)]
+                    for i, child in enumerate(child_list):
+                        props = {'relation_id': k}
+                        if relation.ordered:
+                            props['order'] = i
+                        merge_relationship(self.node, child, type,
+                                           {}, props,
+                                           collision_manager=collision_manager)
+        elif merge_strategy == 'NODE+RELATIONSHIP':
+            parentnames = {p.name: p for p in self.parents}
+            childnames = {p.name: p for p in self.children}
+            parents = []
+            children = []
+            other_parents = []
+            other_children = []
+            for k, (parent_list, relation) in self.predecessors.items():
+                if k in self.identifier_builder and k in parentnames:
+                    if isinstance(parent_list, Collection):
+                        raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
+                    if relation.ordered:
+                        parents += [({'order': i, 'relation_id': k}, p) for i, p in enumerate(parent_list)]
+                    else:
+                        parents += [({'relation_id': k}, p) for i, p in enumerate(parent_list)]
+                elif isinstance(parent_list, Collection):
+                    if relation.ordered:
+                        other_parents.append(({'order': None, 'relation_id': k}, parent_list))
+                    else:
+                        other_parents.append(({'relation_id': k}, parent_list))
+                else:
+                    if relation.ordered:
+                        other_parents += [({'order': i, 'relation_id': k}, p) for i, p in enumerate(parent_list)]
+                    else:
+                        other_parents += [({'relation_id': k}, p) for i, p in enumerate(parent_list)]
                 if k in self.version_on:
-                    version_parents += parent_list
+                    raise NotImplementedError(f"Versioning is not yet implemented: {k} cannot be versioned yet")
+            for k, (child_list, relation) in self.successors.items():
+                if k in self.identifier_builder and k in childnames:
+                    if isinstance(child_list, Collection):
+                        raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
+                    if relation.ordered:
+                        children += [({'order': i, 'relation_id': k}, c) for i, c in enumerate(child_list)]
+                    else:
+                        children += [({'relation_id': k}, c) for i, c in enumerate(child_list)]
+                elif isinstance(child_list, Collection):
+                    if relation.ordered:
+                        other_children.append(({'order': None, 'relation_id': k}, child_list))
+                    else:
+                        other_children.append(({'relation_id': k}, child_list))
+                else:
+                    if relation.ordered:
+                        other_children += [({'order': i, 'relation_id': k}, c) for i, c in enumerate(child_list)]
+                    else:
+                        other_children += [({'relation_id': k}, c) for i, c in enumerate(child_list)]
+                if k in self.version_on:
+                    raise NotImplementedError(f"Versioning is not yet implemented: {k} cannot be versioned yet")
             reltype = 'is_required_by'
-            relparents = {p: (reltype, {'order': 0}, {}) for p in parents}
-            child = self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
-                                           parents=relparents, collision_manager=collision_manager)
-            for i, k, other in others:
-                if other is not None:
-                    merge_relationship(other, child, reltype, {'order': i, 'relation_id': k}, {}, collision_manager=collision_manager)
+            rels = {p: (reltype, True, d, {}) for d, p in parents}
+            rels.update({c: (reltype, False, d, {}) for d, c in children})
+            anti_id_rels = []
+            # make sure that optional relations present in an ID are part of the search
+            # i.e. the absence of an optional relation in the db is itself part of the unique key
+            # anti_id_rels are added for missing optional relations
+            for k in self.identifier_builder:
+                if k not in self.predecessors and k not in self.successors:
+                    p = parentnames.get(k, None)
+                    c = childnames.get(k, None)
+                    if getattr(p, 'is_optional', False):
+                        anti_id_rels.append((p.node.__name__, reltype, True, {'relation_id': p.relation_idname or k}))
+                    if getattr(c, 'is_optional', False):
+                        anti_id_rels.append((c.node.__name__, reltype, False, {'relation_id': p.relation_idname or k}))
+            self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
+                                           id_rels=rels, anti_id_rels=anti_id_rels, collision_manager=collision_manager)
+            for d, others in other_parents:
+                if others is not None:
+                    if isinstance(others, Collection):
+                        with unwind(others, enumerated=True) as (other, i):
+                            if 'order' in d:
+                                d['order'] = i
+                            merge_relationship(other, self.node, reltype, d, {}, collision_manager=collision_manager)
+                        collect(other)
+                    else:
+                        merge_relationship(others, self.node, reltype, d, {}, collision_manager=collision_manager)
+            for d, others in other_children:
+                if others is not None:
+                    if isinstance(others, Collection):
+                        with unwind(others, enumerated=True) as (other, i):
+                            if 'order' in d:
+                                d['order'] = i
+                            merge_relationship(self.node, other, reltype, d, {}, collision_manager=collision_manager)
+                        collect(other)
+                    else:
+                        merge_relationship(self.node, others, reltype, d, {}, collision_manager=collision_manager)
         else:
             ValueError(f"Merge strategy not known: {merge_strategy}")
         if len(version_parents):
-            version_factors = {f: self.neoproperties[f] for f in self.version_on if f in self.factors}
-            set_version(version_parents, ['is_required_by'] * len(version_parents), self.neotypes[-1], child, version_factors)
-        # now the children
-        for k, child_list in successors.items():
-            type = 'is_required_by'
-            if isinstance(child_list, Collection):
-                with unwind(child_list, enumerated=True) as (child, i):
-                    merge_relationship(self.node, child, type,
-                                       {}, {'relation_id': k, 'order': i},
-                                       collision_manager=collision_manager)
-                collect(child)
-            else:
-                for child in child_list:
-                    props =  {'relation_id': k, 'order': 0}
-                    merge_relationship(self.node, child, type, {},
-                                       props, collision_manager=collision_manager)
+            raise NotImplementedError(f"Versioning is not yet implemented: {version_parents} cannot be versioned yet")
 
 
     @classmethod
@@ -466,7 +573,10 @@ class Graphable(metaclass=GraphableMeta):
             return False
         if len(cls.identifier_builder) == 0:
             return False
-        return not any(n.name in cls.identifier_builder for n in cls.parents+cls.children)
+        rels = [x.name if isinstance(x, Multiple) else OneOf(x).name for x in cls.parents+cls.children]
+        if not any(r in cls.factors or r in cls.identifier_builder for r in rels) and rels:
+            raise ValueError(f"{cls} defines an identifier_builder with name(s) which are neither factors, parents, or children")
+        return not any(r in cls.identifier_builder for r in rels)
 
     @classmethod
     def has_rel_identity(cls):
@@ -607,18 +717,25 @@ class Hierarchy(Graphable):
             names = names[0]
         return [cls.from_name(name) for name in names]
 
-    def make_specification(self) -> Tuple[Dict[str, Type[Graphable]], Dict[str, str], Dict[str, Type[Graphable]]]:
+    def make_specification(self) -> Tuple[Dict[str, Type[Graphable]], Dict[str, str], Dict[str, Type[Graphable]], Dict[str, Type[Graphable]]]:
         """
         Make a dictionary of {name: HierarchyClass} and a similar dictionary of factors
         """
         # ordered here since we need to use the first parent as a match point in merge_dependent_node
-        parents = OrderedDict([(getattr(p, 'relation_idname', None) or getattr(p, 'name', None) or p.singular_name, p) for p in self.parents])
-        children = OrderedDict([(getattr(c, 'relation_idname', None) or getattr(c, 'name', None) or c.singular_name, c) for c in self.children])
+        parents = OrderedDict()
+        children = OrderedDict()
+        for inl, outl, mirrorl in ([self.parents, parents, children], [self.children, children, parents]):
+            for p in inl:
+                name = getattr(p, 'relation_idname', None) or getattr(p, 'name', None) or p.singular_name
+                outl[name] = p
+                if isinstance(p, Multiple):
+                    if p.one2one:
+                        mirrorl[name] = p
         factors = {f.lower(): f for f in self.factors}
         specification = parents.copy()
         specification.update(factors)
         specification.update(children)
-        return specification, factors, children
+        return specification, factors, parents, children
 
     @classmethod
     def instantate_nodes(cls, hierarchies=None):
@@ -643,21 +760,20 @@ class Hierarchy(Graphable):
         else:
             self.uses_tables = True
         self.identifier = kwargs.pop(self.idname, None)
-        self.specification, factors, children = self.make_specification()
+        self.specification, factors, parents, children = self.make_specification()
         # add any data held in a neo4j unwind table
-        for k, v in self.specification.items():
-            if k not in kwargs:
-                if isinstance(v, Multiple) and (v.minnumber == 0 or v.notreal):  # i.e. optional
-                    continue
-                if tables is not None:
-                    kwargs[k] = tables.get(tables_replace.get(k, k), alias=False)
-        self._kwargs = kwargs.copy()
+        if tables is not None:
+            for k, v in factors.items():
+                if k not in kwargs:
+                        kwargs[k] = tables.get(tables_replace.get(k, k), alias=False)
         # Make predecessors a dict of {name: [instances of required Factor/Hierarchy]}
         predecessors = {}
         successors = {}
         for name, nodetype in self.specification.items():
             if isinstance(nodetype, Multiple):
-                if (nodetype.minnumber == 0 or nodetype.notreal) and name not in kwargs:
+                if nodetype.notreal and name in kwargs:
+                    raise ValueError(f"{name} is not accepted as an argument for {self} since `notreal` is set for it in {self.__class__.__name__} definition")
+                if (nodetype.minnumber == 0 and name not in kwargs) or nodetype.notreal:
                     continue
             if do_not_create:
                 value = kwargs.pop(name, None)
@@ -672,10 +788,27 @@ class Hierarchy(Graphable):
                                 raise TypeError(f"{name} expects multiple elements")
             else:
                 value = [value]
-            if name in children:
-                successors[name] = value
-            elif name not in factors:
-                predecessors[name] = value
+            try:
+                l = len(value)
+            except TypeError:
+                l = 1
+            nodetype = Multiple.from_any(nodetype)
+            if not isinstance(value, Collection):
+                if not (nodetype.minnumber <= l <= nodetype.maxnumber):
+                    raise ValueError(f"{self.__class__.__name__} requires {nodetype.minnumber} <= count({name}) <= {nodetype.maxnumber}. "
+                                     f"{l} received.")
+                if isinstance(value, Hierarchy):
+                    invalid_types = {v for v in value if isinstance(v, nodetype.node)}
+                    if invalid_types:
+                        raise TypeError(f"{self.__class__.__name__} can only take {nodetype.node.__name__} not {invalid_types}")
+            else:
+                validate_number(value, nodetype.minnumber, nodetype.maxnumber, self.__class__.__name__, nodetype.node.__name__)
+                validate_type(value, nodetype.node.__name__)
+            if name not in factors:
+                if name in children:
+                    successors[name] = value, nodetype
+                if name in parents:
+                    predecessors[name] = value, nodetype
         if len(kwargs):
             raise KeyError(f"{kwargs.keys()} are not relevant to {self.__class__}")
         self.predecessors = predecessors
@@ -706,20 +839,28 @@ class Hierarchy(Graphable):
         except ContextError:
             return
         reltype = 'is_required_by'
-        parents = [getattr(p, 'relation_idname', None) or p.singular_name for p in self.parents if isinstance(p, Multiple) and p.is_optional]
-        children = [getattr(c, 'relation_idname', None) or c.singular_name for c in self.children if isinstance(c, Multiple) and c.is_optional]
+        parents = {p.name: p.one2one for p in self.parents if isinstance(p, Multiple) and p.is_optional}
+        children = {c.name: c.one2one for c in self.children if isinstance(c, Multiple) and c.is_optional}
         for key, item in optionals.items():
+            if isinstance(item, Collection):
+                raise NotImplementedError(f"Cannot attach optional collections yet")
             if key in parents:
                 parent, child = item.node, self.node
+                one2one = parents[key]
             elif key in children:
                 parent, child = self.node, item.node
+                one2one = children[key]
             else:
                 raise KeyError(f"{key} is not a parent or child of {self}")
             merge_relationship(parent, child, reltype, {'order': 0, 'relation_id': key}, {}, collision_manager=collision_manager)
+            if one2one:
+                merge_relationship(child, parent, reltype, {'order': 0, 'relation_id': key}, {}, collision_manager=collision_manager)
 
 def find_branch(*nodes_or_types):
     """
     Given a mix of variables and types along an undirected path (the input order), instantate those types from the graph
-    >>> _, _, l1spectrum, _ = find_branch(fibre, FibreTarget, L1Spectrum, l1file)
+    >>> _, l1spectrum, fibretargetr, _ = find_branch(l1file, L1Spectrum, FibreTarget, fibre)
+    This will find the branch that looks like (l1file)<--(:L1Spectrum)<--(:FibreTarget)<--(fibre)
+    and return the nodes
     """
     return match_branch_node(*[i.__name__ if isinstance(i, type) else i for i in nodes_or_types])
\ No newline at end of file
