@@ -126,26 +126,18 @@ class MatchPatternNode(Statement):
     def to_cypher(self):
         labels = ':'.join(self.labels)
         match = f'({self.out}: {labels} {self.properties})'
-        extras = ', '.join([f'({self.out})<--({p})' for p in self.parents] +
-                            [f'({self.out})-->({c})' for c in self.children])
-        wheres = ' AND '.join([f"({self.out} <> {e})" for e in self.exclude] +
-                              [f"({self.out} <> {e})" for e in self.exclude_collection] +
+        relations = ', '.join([f'({self.out})<--({p})' for p in self.parents] +
+                                [f'({self.out})-->({c})' for c in self.children])
+        extras = ' AND '.join([f"({self.out} <> {e} or {e} is null)" for e in self.exclude] +
+                              [f"all(n in {e} where {self.out} <> n or n is null)" for e in self.exclude_collection] +
                               [f"all(p in {p} where exists( ({self.out})<--(p) ))" for p in self.collection_parents] +
                               [f"all(c in {c} where exists( ({self.out})-->(c) ))" for c in self.collection_children])
-        if len(extras):
-            extras = f',{extras}'
-        if len(wheres):
-            wheres = f' WHERE {wheres}'
-        call_vars = [v for v in self.input_variables if not isinstance(v, CypherData)]
-        if not len(call_vars):
-            return f'WITH * OPTIONAL MATCH {match}{extras}{wheres}'
-        return f"""CALL {{WITH {','.join(map(str, call_vars))}
-         MATCH {match}{extras}{wheres}
-         WITH collect({self.out}) as {self.out}
-         RETURN CASE WHEN SIZE({self.out})=0 THEN [null] ELSE {self.out} END as _{self.out}
-         }}
-         UNWIND _{self.out} as {self.out}
-        """
+        s = f'WITH * OPTIONAL MATCH {match}'
+        if relations:
+            s += f',{relations}'
+        if extras:
+            s += f' WHERE {extras}'
+        return s
 
 class MatchBranchNode(Statement):
     def __init__(self, *nodes_or_labels):
@@ -157,7 +149,7 @@ class MatchBranchNode(Statement):
 
     def to_cypher(self):
         nodes = [f'({o}:{n})' if isinstance(n, str) else f'({o})' for n, o in zip(self.nodes_or_labels, self.output_variables)]
-        path = '--'.join(map('{}'.format, nodes))
+        path = '<-[:is_required_by]-'.join(map('{}'.format, nodes))
         return f'WITH * OPTIONAL MATCH {path}'
 
 class PropertyOverlapError(Exception):
@@ -313,8 +305,9 @@ class MergeRelationship(CollisionManager):
 class MergeDependentNode(CollisionManager):
     def __init__(self, labels: List[str], identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                  properties: Dict[str, Union[str, int, float, CypherVariable]],
-                 parents: List[CypherVariable],
+                 parents: List[CypherVariable], anti_id_rels: Tuple[str, bool, Dict],
                  reltypes: List[str],
+                 reldirs: List[bool], # True if parent, False if child
                  relidentproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                  relproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                  collision_manager='track&flag'):
@@ -330,6 +323,7 @@ class MergeDependentNode(CollisionManager):
             self.relproperties.append(propdict)
             relidentpropins += identpropins
             relpropins += propins
+        self.reldirs = reldirs
         self.parents = parents
         self.outnode = CypherVariable(labels[0])
         self.relvars = [CypherVariable(reltype) for reltype in reltypes]
@@ -338,12 +332,21 @@ class MergeDependentNode(CollisionManager):
         self.reltypes = reltypes
         self.relpropsvars = [CypherVariable(f'{t}_props') for t in reltypes]
         self.colliding_rel_keys = [CypherVariable('colliding_rel_keys') for _ in reltypes]
+        _anti_rel_ids = []
+        anti_id_rel_inputs = []
+        for label, reltype, reldir, props in anti_id_rels:
+            d, anti_id_rel_input = neo4j_dictionary(props)
+            anti_id_rel_inputs += anti_id_rel_input
+            _anti_rel_ids.append((label, reltype, reldir, d))
+        self.anti_id_rels = _anti_rel_ids
         super().__init__(self.outnode, identproperties, properties, collision_manager)
         self.child_holder = CypherVariable('child_holder')
         self.unnamed = CypherVariable('unnamed')
+        self.hashvar = CypherVariable('hash')
         self.input_variables += parents
         self.input_variables += relidentpropins
         self.input_variables += relpropins
+        self.input_variables += anti_id_rel_inputs
         self.output_variables += self.relvars
         self.output_variables += self.dummyrelvars
         self.output_variables += self.relpropsvars
@@ -351,6 +354,7 @@ class MergeDependentNode(CollisionManager):
         self.output_variables.append(self.dummy)
         self.output_variables.append(self.child_holder)
         self.output_variables.append(self.unnamed)
+        self.hidden_variables.append(self.hashvar)
 
     def validate_properties(self):
         super(MergeDependentNode, self).validate_properties()
@@ -371,24 +375,34 @@ class MergeDependentNode(CollisionManager):
     @property
     def merge_statement(self):
         labels = ':'.join(map(str, self.labels))
+        anti_relations = []
         real_relations = []
         temp_relations = []
         test_relations = []
         parent_list = []
-        for i, (parent, reltype, relidentprop, dummyrelvar, relvar) in enumerate(zip(self.parents, self.reltypes, self.relidentproperties, self.dummyrelvars, self.relvars)):
+        for label, reltype, reldir, props in self.anti_id_rels:
+            arrows = ['-' if reldir else '<-', '->' if reldir else '-']
+            rel = f'(:{label}){arrows[0]}[:{reltype} {props}]{arrows[1]}({self.dummy})'
+            anti_relations.append(rel)
+        anti_relations = ' OR '.join(anti_relations)
+        for i, (parent, reltype, reldir, relidentprop, dummyrelvar, relvar) in enumerate(zip(self.parents, self.reltypes, self.reldirs, self.relidentproperties, self.dummyrelvars, self.relvars)):
+            child = f'({self.dummy})'
+            arrows = ['-' if reldir else '<-', '->' if reldir else '-']
+            rel = f'({parent}){arrows[0]}[{dummyrelvar}:{reltype} {relidentprop}]{arrows[1]}'
+            relidentprop[Varname('_query_hash')] = self.hashvar
+            real_rel = f'({parent}){arrows[0]}[{relvar}:{reltype} {relidentprop}]{arrows[1]}'
             if i == 0:
-                child = f'({self.dummy}: {labels} {self.identproperties})'
+                first_child = f'({self.dummy}: {labels} {self.identproperties})'
+                real_child = f'({self.out}: {labels} {self.identproperties})'
             else:
-                child = f'({self.dummy})'
-            rel = f'({parent})-[{dummyrelvar}:{reltype} {relidentprop}]->'
-            real_rel = f'({parent})-[{relvar}:{reltype} {relidentprop}]->'
-            real_child = f'({self.out})'
+                real_child = f'({self.out})'
             real_relations.append(real_rel + real_child)
             temp_relations.append(rel + '(temp)')
             test_relations.append(rel + child)
             parent_list.append(f"{parent}")
-        dct = expand_to_cypher_dict(self.dummy, self.propvar, self.identproperties, *self.parents + self.relidentproperties + self.dummyrelvars)
-        aliases = expand_to_cypher_alias(self.identproperties, *self.parents + self.relidentproperties)
+        r = [{k: v for k, v in r.items() if k != Varname('_query_hash')} for r in self.relidentproperties]
+        dct = expand_to_cypher_dict(self.dummy, self.propvar, self.identproperties, *self.parents + r + self.dummyrelvars)
+        aliases = expand_to_cypher_alias(self.identproperties, *self.parents + r)
         variables = set()
         for p in self.parents:
             variables.add(p)
@@ -407,42 +421,37 @@ class MergeDependentNode(CollisionManager):
                     pass
                 if not isinstance(v, CypherData):
                     variables.add(v)
-        merge_temp_relations = '\n'.join([f'MERGE {r}' for r in temp_relations])
-        merge_final_temp_relation = f"MERGE (temp)-[:TemporaryMerge]->({self.out}: {labels} {self.identproperties})"
+        hashes = [f"apoc.hashing.fingerprinting([{i}, '{rt}', {ri}], {{strategy: 'EAGER'}})" for i, (rt, ri) in enumerate(zip(self.reltypes, r))]
+        hashes.append(f"apoc.hashing.fingerprinting({self.identproperties}, {{strategy: 'EAGER'}})")
+        hashes = "with *, apoc.hashing.fingerprinting([" + ",  ".join(hashes) + f"], {{strategy: 'EAGER'}}) as {self.hashvar}"
+
+        # merge_real_relations = '\n'.join([f'MERGE ({self.out}: {labels} {self.identproperties})'] + [f'MERGE {r}' for r in real_relations])
         merge_real_relations = '\n'.join([f'MERGE {r}' for r in real_relations])
         on_create_rel_returns = ', '.join([f'{relvar}' for relvar in self.relvars])
-        on_match_rel_returns = ', '.join([f'${dummy}[0] as {real}' for dummy, real in zip(self.dummyrelvars, self.relvars)])
+        on_match_rel_returns = ', '.join([f'${dummy} as {real}' for dummy, real in zip(self.dummyrelvars, self.relvars)])
         rel_expansion = expand_to_cypher_alias(self.out, *self.relvars, prefix=f'{self.child_holder}.')
-        optional_match = f'OPTIONAL MATCH {test_relations[0]}'
-        matches = '\n'.join([f'MATCH {t}' for t in test_relations])
-        rel_collection = ', '.join([f'collect({drel}) as {drel}' for rel, drel in zip(self.relvars, self.dummyrelvars)])
-        # include the optional match?
-        collection = f'CALL {{ WITH {", ".join(map(str, variables))}\n' \
-                     f'{optional_match}\n' \
-                     f'{matches}\n' \
-                     f'RETURN collect({self.dummy}) as {self.dummy}, {rel_collection}\n' \
-                     f'}}'
-        condition = f"size({self.dummy}) = 0"
+        exists = ','.join([f'{tr}' for tr in test_relations])
+        notexists_condition = f'WHERE NOT ({anti_relations})' if anti_relations else ''
+        already_exists = f"OPTIONAL MATCH {first_child},{exists} {notexists_condition}"
+        condition = f"{self.dummy} is null"
         iftrue = f"""
-        WITH {aliases}
-        MERGE (temp: TemporaryMerge {{id: $time0}})
-        {merge_temp_relations}
-        {merge_final_temp_relation}
+        WITH {aliases}, $time0 as time0
+        {hashes}
         {merge_real_relations}
-        DETACH DELETE temp
         SET {self.out} += ${self.propvar}
         RETURN {self.out}, {on_create_rel_returns}
         """
-        iffalse = f"RETURN ${self.dummy}[0] as {self.out}, {on_match_rel_returns}"
+        iffalse = f"RETURN ${self.dummy} as {self.out}, {on_match_rel_returns}"
         when = f'CALL apoc.do.when({condition}, "{iftrue}", "{iffalse}", {{ {dct}, time0:time0}}) yield value as {self.child_holder}'
         when += f"\n WITH *, {rel_expansion}"
-        return dedent(f"CALL apoc.lock.nodes({self.parents})\n{collection}\n{when}")
+        return dedent(f"CALL apoc.lock.nodes({self.parents})\n{already_exists}\n{when}")
 
     @property
     def on_match(self):  # remember, we are in a call context
         query = ''
         for i, (r, rprops, colliding_keys) in enumerate(zip(self.relvars+[self.out], self.relpropsvars+[self.propvar], self.colliding_rel_keys+[self.colliding_keys])):
             if self.collision_manager == 'track&flag':
+                raise NotImplementedError(f"This feature has not been adapted yet")
                 if r != self.out:  # handled by the base class above
                     query += dedent(f"""
                         WITH *, [x in apoc.coll.intersection(keys({rprops}), keys(properties({r}))) where ({rprops}[x] is null or {r}[x] is null) or {rprops}[x] <> {r}[x]] as {colliding_keys}
@@ -485,7 +494,7 @@ class MergeDependentNode(CollisionManager):
         aliases = expand_to_cypher_alias(self.out, self.propvar, *self.relvars+self.relpropsvars)
         return dedent(f"""
         // post merge
-        call apoc.do.when(size({self.dummy}) = 0,
+        call apoc.do.when({self.dummy} is null,
         "WITH {aliases}\n{self.on_create}\n RETURN $time0",
         "WITH {aliases}\n{self.on_match}\n RETURN $time0",
         {{ {dct} }}) yield value as {self.unnamed}
@@ -530,6 +539,40 @@ class SetVersion(Statement):
         return '\n'.join(query)
 
 
+class ValidateNumber(Statement):
+    def __init__(self, collection: Collection, mn: int, mx: int, central_typelabel: str, to_join_typelabel: str):
+        self.collection = collection
+        self.mn = mn
+        self.mx = mx
+        self.central_typelabel = central_typelabel
+        self.to_join_typelabel = to_join_typelabel
+        self.dummy = CypherVariable('check')
+        super(ValidateNumber, self).__init__([self.collection], [], [self.dummy])
+
+    def to_cypher(self):
+        return f"WITH *, apoc.util.validatePredicate((size({self.collection}) > {self.mx}) or (size({self.collection}) < {self.mn}), " \
+               f"'{self.central_typelabel} expects {self.mn} <= count({self.to_join_typelabel}) <= {self.mx} not %d', [size({self.collection})]) AS {self.dummy}"
+
+
+
+class ValidateType(Statement):
+    def __init__(self, variable: CypherVariable, typelabel: str):
+        self.variable = variable
+        self.typelabel = typelabel
+        self.dummy = CypherVariable('check')
+        hidden = [self.dummy]
+        if isinstance(self.variable, Collection):
+            self.x = CypherVariable('i')
+            hidden.append(self.x)
+        super(ValidateType, self).__init__([self.variable], [], hidden)
+
+    def to_cypher(self):
+        if isinstance(self.variable, Collection):
+            return f"WITH *, apoc.util.validatePredicate(size([{self.x} in {self.variable} where NOT '{self.typelabel}' in labels({self.x})]) > 0, " \
+                   f"'{self.variable} needs to be a collection of {self.typelabel}', []) as {self.dummy}"
+        return f"WITH *, apoc.util.validatePredicate('NOT {self.typelabel}' in labels({self.variable}), '{self.variable} expects {self.typelabel}', []) as {self.dummy}"
+
+
 def match_node(labels, properties, optional=False):
     query = CypherQuery.get_context()  # type: CypherQuery
     statement = MatchNode(labels, properties, optional)
@@ -582,11 +625,12 @@ def merge_relationship(parent, child, reltype, identproperties, properties, coll
     return statement.out
 
 
-def merge_dependent_node(labels, identproperties, properties, parents, reltypes, relidentproperties, relproperties,
+def merge_dependent_node(labels, identproperties, properties, parents, anti_id_rels,
+                         reltypes, reldirs, relidentproperties, relproperties,
                          collision_manager='track&flag'):
     query = CypherQuery.get_context()  # type: CypherQuery
-    statement = MergeDependentNode(labels, identproperties, properties, parents, reltypes, relidentproperties, relproperties,
-                                   collision_manager)
+    statement = MergeDependentNode(labels, identproperties, properties, parents, anti_id_rels, reltypes, reldirs,
+                                   relidentproperties, relproperties, collision_manager)
     query.add_statement(statement)
     return statement.outnode
 
@@ -597,40 +641,56 @@ def set_version(parents, reltypes, childlabel, child, childproperties):
     query.add_statement(statement)
 
 
+def validate_number(collection, mn, mx, central_typelabel, to_join_typelabel):
+    query = CypherQuery.get_context()  # type: CypherQuery
+    statement = ValidateNumber(collection, mn, mx, central_typelabel, to_join_typelabel)
+    query.add_statement(statement)
+
+def validate_type(collection, typelabel):
+    query = CypherQuery.get_context()  # type: CypherQuery
+    statement = ValidateType(collection, typelabel)
+    query.add_statement(statement)
+
+
 def merge_node(labels, identproperties, properties=None,
-               parents: Dict[CypherVariable, Union[Tuple[str, Optional[Dict], Optional[Dict]], str]] = None,
+               id_rels: Dict[CypherVariable, Union[Tuple[str, bool, Optional[Dict], Optional[Dict]], str]] = None,
+               anti_id_rels: List[Tuple[str, bool, Dict]] = None,
                versioned_label=None,
                versioned_properties=None,
                collision_manager='track&flag') -> CypherVariable:
     if properties is None:
         properties = {}
-    if parents is None:
-        parents = {}
-    parent_list = []
+    if id_rels is None:
+        id_rels = {}
+    if anti_id_rels is None:
+        anti_id_rels = {}
+    rel_list = []
+    reldir_list = []
     reltype_list = []
     relidentproperties_list = []
     relproperties_list = []
-    for parent, reldata in parents.items():
+    for rel, reldata in id_rels.items():
         if isinstance(reldata, str):
-            reldata = [reldata]
-        parent_list.append(parent)
+            reldata = [True, *reldata]
+        rel_list.append(rel)
         reltype_list.append(reldata[0])
+        reldir_list.append(reldata[1])
         if len(reldata) > 1:
-            relidentproperties_list.append(reldata[1])
+            relidentproperties_list.append(reldata[2])
         else:
             relidentproperties_list.append({})
         if len(reldata) > 2:
-            relproperties_list.append(reldata[2])
+            relproperties_list.append(reldata[3])
         else:
             relproperties_list.append({})
-    if len(parents):
-        node = merge_dependent_node(labels, identproperties, properties, parent_list,
-                                    reltype_list, relidentproperties_list, relproperties_list,
+    if len(id_rels) or len(anti_id_rels):
+        node = merge_dependent_node(labels, identproperties, properties, rel_list, anti_id_rels,
+                                    reltype_list, reldir_list, relidentproperties_list, relproperties_list,
                                     collision_manager)
     else:
         node = merge_single_node(labels, identproperties, properties, collision_manager)
     if versioned_label is not None:
         if versioned_properties is None:
             versioned_properties = {}
-        set_version(parent_list, reltype_list, versioned_label, node, versioned_properties)
+        set_version(rel_list, reltype_list, versioned_label, node, versioned_properties)
     return node
